package main

// CLAPTRAP_USAGE_JSON=true ./leg | makeclapman man.txt

import (
	// v1	"bufio"
	"bufio"
	"fmt"
	"io"
	"os"
	"regexp"
	"strconv"
	"strings"

	// Flags parsing

	"ser1.net/claptrap/v4"

	// Mainly for manpage generation
	"ser1.net/claphelp/v2"
	. "ser1.net/legume"
	"ser1.net/legume/compact"
	"ser1.net/legume/csv"
	"ser1.net/legume/detail"
	"ser1.net/legume/details"
	"ser1.net/legume/fs"
	"ser1.net/legume/json"
	"ser1.net/legume/stream"
	"ser1.net/legume/vim"
	"ser1.net/legume/xml"
)

// XXX: cross references, e.g. main.go#9. When found, include in detailed output and as x-refs in JSON
// TODO: Allow for escaping special character's, e.g. \: shouldn't be parsed, but should be rendered as just a colon
// TODO: Setting GOMAXPROCS=1 doubles the performance. Is IO threaded?

var Version = `v1.4.2`

func main() {
	reg := claptrap.Command("leg", "legume is an issue tracker using code comments to track tickets.")
	claptrap.Usage = claphelp.PrettyUsage

	// TODO (A) add changelog format, for piping into changelogs
	reg.Add("--format", "-f", []string{"compact", "detail", "csv", "xml", "json", "vim"}, "output format")
	reg.Add("--details...", "-d", 0, "print details about an indexed issue(s) (the flag may be omitted). This is the default action for int arguments.")
	reg.Add("--include...", "-p", "", "filters for filepath matching EXPR")
	reg.Add("--exclude...", "-P", "", "filters for filepath NOT matching EXPR")
	reg.Add("--alias-extension...", "-a", "", "extension alias, e.g. -a cpp%%c") // TODO: Add unit tests for Alias component:ui
	reg.Add("--keyword...", "-w", "", "additional flagword; e.g. -w Hack")       // TODO: Implement & add unit tests for Keywords component:ui
	reg.Add("--output", "-o", "", "output file (if not present, prints to stdout)")
	reg.Add("--help", "-h", false, "print help")
	reg.Add("--version", "-V", false, "print program version")
	reg.Add("--diff", "read from stdin as a diff")
	reg.Add("items...", ".", "items to print details about, or paths (files or dirs) to investigate")
	// REALLY simple persistent argument code
	arguments := os.Args[1:]
	legrc := ".legrc"
	if _, err := os.Stat(legrc); err == nil {
		cf, err := os.Open(legrc)
		defer cf.Close()
		if err == nil {
			config := bufio.NewScanner(cf)
			for config.Scan() {
				argValue := strings.SplitN(config.Text(), " ", 2)
				arguments = append(arguments, argValue...)
			}
		} else {
			fmt.Fprintf(os.Stderr, "skipping %s due to open error: %s\n", legrc, err)
		}
	}
	reg.Parse(arguments)
	if reg.Bool("help") {
		claptrap.Usage()
		os.Exit(0)
	}
	if reg.Bool("version") {
		fmt.Printf("legume %s\n", Version)
		os.Exit(0)
	}

	aliases, ers := makeAlias(remove(reg.Strings("alias-extension"), ""))
	for _, err := range ers {
		fmt.Printf(err.Error())
	}
	ticketNumbers := make([]int, 0)
	items := remove(reg.Strings("items"), "")
	sources := make([]string, 0, len(items))
	for _, s := range items {
		if _, err := os.Stat(s); err == nil { // Files take precidence
			sources = append(sources, s)
		} else if v, err := strconv.Atoi(s); err == nil { // But if it's not a file, and it's a number, then assume it's a ticket
			ticketNumbers = append(ticketNumbers, v)
		}
	}
	if len(sources) == 0 {
		sources = append(sources, ".")
	}
	if reg.Args["details"].IsSet() {
		for _, d := range reg.Ints("details") {
			ticketNumbers = append(ticketNumbers, d)
		}
	}
	// FIXME placeholder for diff work -- be smarter about the files when isDiff
	if reg.Bool("diff") {
		sources = []string{}
	}
	inEx := makeIncExc(remove(reg.Strings("include"), ""), remove(reg.Strings("exclude"), ""), sources)
	var sourcePipe <-chan SourceFile
	var errPipe <-chan error
	isDiff := reg.Bool("diff")
	if isDiff {
		sourcePipe, errPipe = stream.MakeSource(aliases, inEx)
	} else {
		// If the arguments are only integers, and files don't exist by those names, assume they're ticket numbers
		sourcePipe, errPipe = fs.MakeSource(sources, aliases, inEx)
	}
	var output io.Writer
	outFile := reg.String("output")
	if outFile != "" {
		fout, err := os.Create(outFile)
		if err != nil {
			fmt.Println(err.Error())
			return
		}
		defer fout.Close()
		output = fout
	} else {
		output = os.Stdout
	}
	var toString func(i []Issue, out io.Writer)
	if len(ticketNumbers) > 0 {
		toString = func(i []Issue, out io.Writer) {
			details.ToString(ticketNumbers, i, out)
		}
	} else {
		switch reg.String("format") {
		case "vim":
			toString = vim.ToString
		case "csv":
			toString = csv.ToString
		case "compact":
			toString = compact.ToString
		case "detail":
			toString = detail.ToString
		case "json":
			toString = json.ToString
		case "xml":
			toString = xml.ToString
		default:
			toString = compact.ToString
		}
	}
	// TODO: Filter by keyword, e.g. type=BUG
	filter := func(i Issue) bool {
		return true
	}
	if isDiff {
		filter = func(i Issue) bool {
			return i.Change != ACTIVE
		}
	}
	// TODO stream issues, and avoid the array
	issues := findIssues(sourcePipe, errPipe, reg.Strings("keyword"), filter)
	toString(issues, output)
}

func remove[T comparable](arry []T, item T) []T {
	if len(arry) == 0 {
		return arry
	}
	for len(arry) > 0 && arry[0] == item {
		arry = arry[1:]
	}
	if len(arry) == 0 {
		return arry
	}
	for len(arry) > 0 && arry[len(arry)-1] == item {
		arry = arry[:len(arry)-1]
	}
	if len(arry) == 0 {
		return arry
	}
	i, idx := 1, 1
	for ; i < len(arry); i, idx = i+1, idx+1 {
		if arry[i] == item {
			idx = idx - 1
			continue
		}
		if i != idx {
			arry[idx] = arry[i]
		}
	}
	return arry[0 : len(arry)-(i-idx)]
}

// Builds an alias map from unparsed arguments. The key of the resulting map is
// the new suffix; the value is the associated suffix to use for comment style
func makeAlias(alias []string) (map[string]string, []error) {
	rv := make(map[string]string)
	erv := make([]error, 0)
	for _, f := range alias {
		if f == "" {
			continue
		}
		parts := strings.Split(f, "%")
		if len(parts) == 2 {
			if len(parts[0]) == 0 {
				erv = append(erv, fmt.Errorf("bad alias format \"%s\"; format should be <new>%%<old>", f))
			} else if len(parts[1]) == 0 {
				erv = append(erv, fmt.Errorf("bad alias format \"%s\"; format should be <new>%%<old>", f))
			}
			if len(parts[0]) != 0 && len(parts[1]) != 0 {
				rv[parts[0]] = parts[1]
			}
		} else {
			erv = append(erv, fmt.Errorf("bad alias format \"%s\"; format should be <new>%%<old>", f))
		}
	}
	return rv, erv
}

// Builds a file name filter.  The resulting function takes a file name and
// returns true if the file name matches the filterPath regexp and does not match
// the filterNotPath regexp. If no patterns were provided, match everything. If an
// include pattern was provided, and source files were provided, include the starting
// sourceFiles (even if they may not match the pattern) -- this allows directory matching.
func makeIncExc(filterPath, filterNotPath, sourceFiles []string) func(s string) bool {
	if len(filterPath) == 0 && len(filterNotPath) == 0 {
		return func(i string) bool {
			return true
		}
	}
	inclMatches := make([]*regexp.Regexp, 0, len(filterPath))
	for _, m := range filterPath {
		re, err := regexp.Compile(m + "$")
		if err != nil {
			fmt.Printf("bad path expression \"%s\". Must be a regular expression.\n", m)
		} else {
			inclMatches = append(inclMatches, re)
		}
	}
	if len(inclMatches) > 0 {
		// Add the starting directory argument
		for _, m := range sourceFiles {
			re, err := regexp.Compile(regexp.QuoteMeta(m) + "$")
			if err != nil {
				fmt.Printf("bad path expression \"%s\". Must be a regular expression.\n", m)
			} else {
				inclMatches = append(inclMatches, re)
			}
		}
	}
	exclMatches := make([]*regexp.Regexp, 0, len(filterNotPath))
	for _, m := range filterNotPath {
		re, err := regexp.Compile(m + "$")
		if err != nil {
			fmt.Printf("bad path expression \"%s\". Must be a regular expression.\n", m)
		} else {
			exclMatches = append(exclMatches, re)
		}
	}
	return func(i string) bool {
		var incl bool
		// If user did not specify includes, include everything by default
		incl = len(inclMatches) == 0
		for _, x := range inclMatches {
			if x.MatchString(i) {
				incl = true
				break
			}
		}
		for _, x := range exclMatches {
			if x.MatchString(i) {
				incl = false
				break
			}
		}
		return incl
	}
}

type filterFunc func(i Issue) bool
