package main

import (
	"reflect"
	"testing"
)

// Test_makeAlias tests file suffix aliasing, and exercised the source file suffix returning the correct language
func Test_makeAlias(t *testing.T) {
	tests := []struct {
		name        string
		args        []string
		wantAliases map[string]string
		wantErrors  int
	}{
		{"good args", []string{"a%b", "c%d"}, map[string]string{"a": "b", "c": "d"}, 0},
		{"empty args", []string{}, map[string]string{}, 0},
		{"too many parts", []string{"a%b", "c%d%e"}, map[string]string{"a": "b"}, 1},
		{"no key", []string{"%b"}, map[string]string{}, 1},
		{"no value", []string{"a%"}, map[string]string{}, 1},
		{"no text", []string{"%"}, map[string]string{}, 1},
		{"no sep", []string{"ab"}, map[string]string{}, 1},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			aliases, errors := makeAlias(tt.args)
			if !reflect.DeepEqual(aliases, tt.wantAliases) {
				t.Errorf("makeAlias() %s aliases = %v, wantAliases %v", tt.name, aliases, tt.wantAliases)
			}
			numErrors := len(errors)
			if numErrors != tt.wantErrors {
				errMsgs := ""
				for _, e := range errors {
					errMsgs = errMsgs + "\"" + e.Error() + "\" "
				}
				t.Errorf("makeAlias() %s errors = %s, wanted %d errors", tt.name, errMsgs, tt.wantErrors)
			}
		})
	}
}

// Test_makeIncExc tests that file inclusion/exclusion matches correctly filters a set of source files
func Test_makeIncExc(t *testing.T) {
	tests := []struct {
		name string
		path string
		incl []string
		excl []string
		want bool
	}{
		{"no args", "a.c", []string{}, []string{}, true},
		{"match path", "a.c", []string{"\\.c$"}, []string{}, true},
		{"match path 2", "a.c", []string{"^a.*"}, []string{}, true},
		{"no match path", "a.d", []string{"\\.c$"}, []string{}, false},
		{"match not path", "a.c", []string{}, []string{"\\.c$"}, false},
		{"match not path 2", "a.d", []string{}, []string{"\\.c$"}, true},
		{"not path overrides", "a.c", []string{"\\.c$"}, []string{"\\.c$"}, false},
		{"not path overrides 2", "a.c", []string{"^a"}, []string{"\\.c$"}, false},
		{"match both", "a.c", []string{"\\.c$"}, []string{"\\.d$"}, true},
		{"match both 2", "a.d", []string{"\\.c$"}, []string{"\\.d$"}, false},
		{"match path no re", "./cmd/foo/bar/abc.java", []string{"cmd"}, []string{}, false},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if ff := makeIncExc(tt.incl, tt.excl, []string{}); ff(tt.path) != tt.want {
				t.Errorf("%s: %q => %v, want %v", tt.name, tt.path, ff(tt.path), tt.want)
			}
		})
	}
}

/* NOT YET IMPLEMENTED
// Test_makeFilter tests the filtering parsed issues by KEYWORD=VALUE
func Test_makeFilter(t *testing.T) {
	issue := Issue{Tags: map[string]string{"tag": "tagval"}, Description: "desc"}
	tests := []struct {
		name  string
		issue Issue
		filterFor []string
		filterNot []string
		filterDesc []string
		want  bool
	}{
		{"match tag", issue, []string{"tag:tagval"}, []string{""}, []string{""}, true},
		{"match tag", issue, []string{"tag:tagval"}, []string{""}, []string{""}, true},
		{"match desc", issue, []string{""}, []string{""}, []string{"desc"}, true},
		{"match not tag", issue, []string{""}, []string{"tag:tagval"}, []string{""}, false},
		{"match not tag 2", issue, []string{""}, []string{"gat:tagval"}, []string{""}, true},
		{"match tag & desc", issue, []string{"tag:tagval"}, []string{""}, []string{"desc"}, true},
		{"match desc w/ mismatch tag", issue, []string{"gat:fgoo"}, []string{""}, []string{"desc"}, true},
		{"match tag w/ mismatch desc", issue, []string{"tag:tagval"}, []string{""}, []string{"csec"}, true},
		{"not tag overrides tag", issue, []string{"tag:tagval"}, []string{"tag:tagval"}, []string{""}, false},
		{"not tag overrides desc", issue, []string{""}, []string{"tag:tagval"}, []string{"desc"}, false},
		{"not tag & not desc overrides tag & desc", issue, []string{"tag:tagval"}, []string{"tag:tagval"}, []string{"desc"}, false},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if got := makeFilter(tt.args.a)(tt.issue); !reflect.DeepEqual(got, tt.want) {
				t.Errorf("makeFilter() %s = %v, want %v", tt.name, got, tt.want)
			}
		})
	}
}
*/
