package fs

import (
	"errors"
	"os"
	"path/filepath"
	"strings"

	. "ser1.net/legume"
	"ser1.net/legume/parse"
	"ser1.net/legume/reader"
	"ser1.net/legume/todo"
)

// SourceFiles from the filesystem.
// Includes all files with recognized file suffixes, or if the suffix is a key
// in the map then the file is included as if it were of the type returned as
// the value.  Uses the supplied function to determine whether or not to
// include the file by passing in the relative path to the file as the function
// argument.
func MakeSource(start []string, a map[string]string, inEx func(s string) bool) (<-chan SourceFile, <-chan error) {
	csf := make(chan SourceFile)
	cer := make(chan error)
	go func() {
		makeSourceDir(a, inEx, start, csf, cer)
		close(csf)
		close(cer)
	}()
	return csf, cer
}

func makeSourceDir(a map[string]string, f func(s string) bool, dirs []string, cfs chan<- SourceFile, cer chan<- error) {
	for _, dir := range dirs {
		filepath.Walk(dir, func(path string, info os.FileInfo, err error) error {
			// Processes files.  Directories are walked but not processed.
			if err != nil {
				cer <- err
				return nil
			}
			if !f(path) {
				if info.IsDir() {
					return filepath.SkipDir
				}
				return nil
			}
			// The rest of the func only applies to files, so return
			if info.IsDir() {
				return nil
			}
			ext := filepath.Ext(info.Name())
			if len(ext) < 2 {
				return nil
			}
			if strings.ToLower(info.Name()) == "todo.txt" {
				f, err := os.Open(path)
				if err != nil {
					cer <- err
					return nil
				}
				k := todo.New(f, path)
				cfs <- &k
				return nil
			}
			ext = ext[1:]
			_, ok := parse.Languages[ext]
			_, ok2 := a[ext]
			if !ok && !ok2 {
				return nil
			}
			if !ok {
				_, ok = parse.Languages[a[ext]]
				if !ok {
					cer <- errors.New("invalid alias " + ext + " : " + a[ext])
					return nil
				}
				ext = a[ext]
			}
			f, err := os.Open(path)
			if err != nil {
				cer <- err
				return nil
			}
			k := reader.New(f, ext, path)
			cfs <- &k
			return nil
		})
	}
}
