package parse

import (
	"bufio"
	"errors"
	"regexp"
	"strings"
	"time"

	. "ser1.net/legume"
)

const (
	// IN ORDER
	// key:2017-04-30
	// key:value
	// @context
	// +project
	// 2017-04-30
	// (A)
	//TOKENS = "(?:\\s|^)(?:(?:\\w+:\\d{4}-\\d{2}-\\d{2})|(?:\\w+:\\w+)|(?:@\\w+)|(?:\\+\\w+)|(?:\\d{4}-\\d{2}-\\d{2})|(?:\\([A-Z]+\\))(?:\\s|$))"
	TOKENS0 = "\\b((\\w+:\\d{4}-\\d{2}-\\d{2})|(\\w+:\\w+)|(\\d{4}-\\d{2}-\\d{2}))\\b"
	TOKENS1 = "(\\s|^)((@\\w+)|(\\+\\w+))\\b"
	TOKENS2 = "(\\s|^)(\\([A-Z]+\\))(\\s|$)"
)

var flagWords []string = []string{"todo", "fixme", "xxx", "bug"}
var kwMatch *regexp.Regexp
var tkMatch []*regexp.Regexp

func init() {
	SetKeywords([]string{})
	tkMatch = []*regexp.Regexp{
		regexp.MustCompile(TOKENS0),
		regexp.MustCompile(TOKENS1),
		regexp.MustCompile(TOKENS2),
	}
}

func SetKeywords(ks []string) (err error) {
	matchString := "^\\s*((?i)"
	good := regexp.MustCompile("^\\w+$")
	var prepend bool
	for i := 0; i < len(flagWords); i++ {
		if prepend {
			matchString += "|"
		} else {
			prepend = true
		}
		matchString += flagWords[i]
	}
	for i := 0; i < len(ks); i++ {
		if !good.MatchString(ks[i]) {
			return errors.New("keywords must be only words!")
		}
		matchString += "|" + ks[i]
	}
	matchString += ")[:]? "
	kwMatch, err = regexp.Compile(matchString)
	return err
}

func parseIssues(c comment) []Issue {
	lines := bufio.NewScanner(strings.NewReader(c.String))
	issues := make([]Issue, 0)
	var cont bool
	for lines.Scan() {
		line := lines.Text()
		if issue, got := parseIssue(line); got {
			issue.Row = c.Line
			issue.Change = c.Change
			issues = append(issues, issue)
			cont = true
		} else {
			// FIXME: If there's a blank line, we shouldn't concat comments
			if cont {
				if len(strings.TrimSpace(line)) > 0 {
					i := issues[len(issues)-1]
					i.Description += "\n" + strings.TrimSpace(line)
					issues[len(issues)-1] = i
				} else {
					cont = false
				}
			}
		}
	}
	return issues
}

/*
 * simple issue parsing
 *
 * an issue is a flagword, followed by : , followed by some description and
 * ended optionally by some tags (No t/f? End by whiteline or eof or
 * another TODO).
 *
 * tags can be placed *before* description, too
 */
func parseIssue(l string) (Issue, bool) {
	kws := kwMatch.FindStringSubmatch(l)
	if len(kws) == 0 {
		return Issue{}, false
	}
	l = kwMatch.ReplaceAllString(l, "")
	var res []string
	rest := l
	for _, match := range tkMatch {
		res = append(res, match.FindAllString(l, -1)...)
		rest = match.ReplaceAllString(rest, "")
	}
	tags := make(map[string]string)
	contexts := make([]string, 0)
	projects := make([]string, 0)
	var date *time.Time
	var pri string
	for _, m := range res {
		m = strings.TrimSpace(m)
		switch m[0] {
		case '@': // context
			contexts = append(contexts, m[1:])
		case '+': // project
			projects = append(projects, m[1:])
		case '0', '1', '2', '3', '4', '5', '6', '7', '8', '9': // date
			// What do you do with an error?
			d, _ := time.Parse("2006-01-02", m)
			date = &d
		case '(': // priority
			pri = m[1 : len(m)-1]
		default: // key:value or key:date }
			ks := strings.Split(m, ":")
			tags[ks[0]] = ks[1]
		}
	}
	return Issue{
		Row:         -1,
		Description: strings.TrimSpace(rest),
		Tags:        tags,
		Contexts:    contexts,
		Projects:    projects,
		Date:        date,
		Priority:    pri,
		Path:        "PATH",
		Change:      ACTIVE,
		Type:        strings.ToUpper(kws[1]),
	}, true
}
