package parse

import (
	"reflect"
	"regexp"
	"testing"
	"time"

	. "ser1.net/legume"
)

func TestSetKeywords(t *testing.T) {
	tests := []struct {
		name    string
		args    []string
		wantErr bool
		wantRe  string
	}{
		{"default", []string{}, false, regexp.MustCompile("^\\s*(?i:todo|fixme|xxx)[:]? ").String()},
		{"bad", []string{"("}, true, ""},
		{"bad 2", []string{" "}, true, ""},
		{"bad 3", []string{"a-b"}, true, ""},
		{"good", []string{"foo", "bar"}, false, regexp.MustCompile("^\\s*(?i:todo|fixme|xxx|foo|bar)[:]? ").String()},
		{"bad 4", []string{"foo", "-"}, true, ""},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if err := SetKeywords(tt.args); (err != nil) != tt.wantErr {
				t.Errorf("SetKeywords() %s error = %v, wantErr %v", tt.name, err, tt.wantErr)
				if kwMatch != nil {
					t.Errorf("SetKeywords() %s re = %v, wantRe %v", tt.name, kwMatch.String(), tt.wantRe)
				} else {
					t.Errorf("SetKeywords() %s re = %v, wantRe %v", tt.name, kwMatch, tt.wantRe)
				}
			}
		})
	}
}

func Test_parseIssues(t *testing.T) {
	tests := []struct {
		name string
		args comment
		want []Issue
	}{
		{"no match", comment{Line: 0, String: ` 00000 `}, []Issue{}},
		{"match no colon", comment{Line: 0, String: ` TODO blah`}, []Issue{
			Issue{
				Path:        "PATH",
				Row:         0,
				Description: "blah",
				Contexts:    make([]string, 0),
				Projects:    make([]string, 0),
				Tags:        make(map[string]string),
				Type:        "TODO",
			},
		}},
		{"match colon", comment{Line: 1, String: ` fixme: blah blah`}, []Issue{
			Issue{
				Path:        "PATH",
				Row:         1,
				Description: "blah blah",
				Contexts:    make([]string, 0),
				Projects:    make([]string, 0),
				Tags:        make(map[string]string),
				Type:        "FIXME",
			},
		}},
		{"missing space", comment{Line: 2, String: ` fixme:blah blah`}, []Issue{}},
		{"space not necessary", comment{Line: 9, String: `fixme: blah blah blah`}, []Issue{
			Issue{
				Path:        "PATH",
				Row:         9,
				Description: "blah blah blah",
				Contexts:    make([]string, 0),
				Projects:    make([]string, 0),
				Tags:        make(map[string]string),
				Type:        "FIXME",
			},
		}},
		{"contexts", comment{Line: 3, String: `xxx: blah blah @one @two`}, []Issue{
			Issue{
				Path:        "PATH",
				Row:         3,
				Description: "blah blah",
				Contexts:    []string{"one", "two"},
				Projects:    make([]string, 0),
				Tags:        make(map[string]string),
				Type:        "XXX",
			},
		}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if got := parseIssues(tt.args); !reflect.DeepEqual(got, tt.want) {
				t.Errorf("parseIssues() %s = %#v, want %#v", tt.name, got, tt.want)
			}
		})
	}
}

func Test_parseIssue(t *testing.T) {
	d, _ := time.Parse("2006-01-02", "2017-01-02")
	date := &d
	mta := make([]string, 0)
	mtm := make(map[string]string)
	tests := []struct {
		name  string
		args  string
		want  Issue
		wantm bool
	}{
		{"no match", "xx B)", Issue{}, false},
		{"no match priorities", "xxx (A)(B)", Issue{
			Path:        "PATH",
			Row:         -1,
			Description: "(A)(B)",
			Contexts:    mta,
			Projects:    mta,
			Tags:        mtm,
			Type:        "XXX",
		}, true},
		{"match series contexts", "todo @a @b @c @d ok", Issue{
			Path:        "PATH",
			Row:         -1,
			Description: "ok",
			Contexts:    []string{"a", "b", "c", "d"},
			Projects:    mta,
			Tags:        mtm,
			Type:        "TODO",
		}, true},
		{"match series projects", "todo +a +b ok +c +d ", Issue{
			Path:        "PATH",
			Row:         -1,
			Description: "ok",
			Projects:    []string{"a", "b", "c", "d"},
			Contexts:    mta,
			Tags:        mtm,
			Type:        "TODO",
		}, true},
		// This is messy. The amount of code to make this one work right is annoying.
		// Maybe leave it as an edge case -- behavior undefined.
		//{"one priority", "todo ok (A) (B)", Issue{}, false},
		{"everything", "xxx: 2017-01-02 (A) @one +abc blah blah @two +def tag:value other:tag due:2017-03-03", Issue{
			Path:        "PATH",
			Row:         -1,
			Description: "blah blah",
			Contexts:    []string{"one", "two"},
			Projects:    []string{"abc", "def"},
			Date:        date,
			Priority:    "A",
			Tags:        map[string]string{"tag": "value", "other": "tag", "due": "2017-03-03"},
			Type:        "XXX",
		}, true},
		{"regexp test", "xxx: a+b (a) key: :value a@b 2017", Issue{
			Path:        "PATH",
			Row:         -1,
			Description: "a+b (a) key: :value a@b 2017",
			Contexts:    mta,
			Projects:    mta,
			Tags:        mtm,
			Type:        "XXX",
		}, true},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, match := parseIssue(tt.args)
			if !reflect.DeepEqual(got, tt.want) {
				t.Errorf("parseIssues() %s = %#v, want %#v", tt.name, got, tt.want)
			}
			if !reflect.DeepEqual(match, tt.wantm) {
				t.Errorf("parseIssues() %s = %v, want %v", tt.name, match, tt.wantm)
			}
		})
	}
}
