package stream

import (
	"errors"
	"io"
	"os"
	"path/filepath"
	"strings"

	dp "github.com/waigani/diffparser"
	. "ser1.net/legume"
	"ser1.net/legume/parse"
)

// TODO (A) Include the time stamp of last change; if item was removed, use old version's time; if added, use new version's @diff
type diffSource struct {
	// file is the original diffed file reference
	file *dp.DiffFile
	// typ is the processed legume (source) type of the file
	typ string
	// text is the (next) scanned line
	text string
	// oline is the index of the "Original" diff line currently being processed
	oline int
	// nline is the index of the "New" diff line currently being processed
	nline int
	// hunk is the index of the hunk currently being processed
	hunk int
	// line is the sourcecode line number; this will be relative to the Original or New, depending on hunk
	line int
	// mode is the diff change type (add/remove/change)
	mode dp.DiffLineMode
}

// Scan looks through the diff and sets up each Text() line.
// It first iterates over all of the original lines, looking for changes.
// It then scans over all of the new lines, looking for changes.
// When it's processed all of the original & new lines, it skips to the next hunk.
// There will be a different diffSource for each file.
func (d *diffSource) Scan() bool {
RETRY:
	switch {
	case len(d.file.Hunks[d.hunk].OrigRange.Lines) > d.oline:
		lines := d.file.Hunks[d.hunk].OrigRange.Lines[d.oline]
		d.text = lines.Content
		d.line = lines.Number
		d.mode = lines.Mode
		d.oline++
	case len(d.file.Hunks[d.hunk].NewRange.Lines) > d.nline:
		lines := d.file.Hunks[d.hunk].NewRange.Lines[d.nline]
		d.text = lines.Content
		d.line = lines.Number
		d.mode = lines.Mode
		d.nline++
	case len(d.file.Hunks) > d.hunk+1: // No more lines in this hunk, so move to next hunk
		d.oline, d.nline = 0, 0
		d.hunk++
		goto RETRY
	default:
		return false
	}
	return true
}

func (f diffSource) Text() string {
	return f.text
}

func (f diffSource) Line() int {
	return f.line
}

func (f diffSource) Type() string {
	return f.typ
}

func (f diffSource) Change() int {
	if f.mode == dp.UNCHANGED {
		return 0
	} else {
		return int(f.mode) + 1
	}
}

func (f diffSource) Path() string {
	return f.file.OrigName
}

func MakeSource(aliases map[string]string, inEx func(s string) bool) (<-chan SourceFile, <-chan error) {
	strm := make(chan SourceFile)
	errs := make(chan error)
	go processIO(aliases, inEx, strm, errs)
	return strm, errs
}

func processIO(aliases map[string]string, inEx func(s string) bool, sfs chan<- SourceFile, ers chan<- error) {
	defer close(sfs)
	defer close(ers)
	// The diff parsing library parses a full diff as a string,
	// so there's no way to not read the whole diff into memory.
	bts, err := io.ReadAll(os.Stdin)
	if err != nil {
		ers <- err
		return
	}
	diff, err := dp.Parse(string(bts))
	if err != nil {
		ers <- err
		return
	}
	for _, file := range diff.Files {
		fname := getName(file.OrigName)
		ext := filepath.Ext(fname)
		if len(ext) < 2 {
			continue // Skip files with no extension
		}
		ext = ext[1:]
		_, ok := parse.Languages[ext]
		_, ok2 := aliases[ext]
		if !ok && !ok2 {
			continue // Skip files we don't recognize
		}
		if !ok {
			_, ok = parse.Languages[aliases[ext]]
			if !ok {
				ers <- errors.New("invalid alias " + ext + " : " + aliases[ext])
				continue
			}
			ext = aliases[ext]
		}
		ds := diffSource{
			file: file,
			typ:  ext,
		}
		sfs <- &ds
	}
}

// Find the hidden file name in the diff header string
// cmd/leg/main.go Wed Apr 26 14:58:54 2017 -0400
func getName(s string) string {
	parts := strings.Split(s, "\t")
	return parts[0]
}

var diffError error

// Given a list of DiffLines, find the next one that's a change and reposition
// the slice to start at that.
func advance(ls []*dp.DiffLine) []*dp.DiffLine {
	for {
		if ls[0].Mode == dp.ADDED || ls[0].Mode == dp.REMOVED {
			return ls
		}
		if len(ls) > 1 {
			ls = ls[1:]
		} else {
			return ls[:0]
		}
	}
}
